#include <vector>
#include <iostream>
#include <string.h>

extern "C" {
  extern int lbfgs(long* n, long* m, double* x, double* f, double* g, 
		    long* diagco, double* diag, long* iprint, double* eps, 
		    double* xtol, double* w, long* iflag);

  extern int lbfgsb(long *n, long *m, double *x, double *l, double *u, long *nbd, 
		     double *f, double *g, double *factr, double *pgtol, double *wa, long *iwa, 
		     char *task, long *iprint, char *csave, long *lsave, long *isave, double *dsave, short task_len, short csave_len); 
}

using namespace std;

class LBFGS {

private:
  long n;
  long m;
  long iflag;
  double *l;
  double *u;
  long   *nbd;
  long type; 
  char   task[64];
  char   csave[64];
  long   lsave[4];
  long   isave[64];
  double dsave[32];

  std::vector <double> diag;
  std::vector <double> w;
  std::vector <long>   iw;

public:

  // these valuables are fixed.
  LBFGS(): n(0), m(5), iflag(0), l(0), u(0), nbd(0), type(-1) {};  
  ~LBFGS() {};

  const char *what () { return task; }

  int init (long _n, long _m, double *_l = 0, double *_u = 0, long *_nbd = 0)
  {
    n = _n;
    m = _m;

    if (_l && _u && _nbd) {
      l = _l;
      u = _u;
      nbd = _nbd;
      w.resize ((2 * m + 4) * n + 12 * m * m + 12 * m);
      iw.resize (3 * n);
      iflag = 0;
      type = 1; // LBFGS-B
    } else {
      iflag = 0;
      w.resize (n * (2 * m + 1) + 2 * m);
      diag.resize (n);
      type = 0; // LBFGS
    }

    return 0;
  }
  
  int optimize (double *x, double *f, double *g)   // vector, object, gradient
  {
    switch (type) {

    case 0: {
      long iprint[] = {-1, 0};
      double eta  = 1e-5;
      double xtol = 1e-5;
      long diagco = 0;
      lbfgs(&n, &m, x, f, g, &diagco, &diag[0], iprint, &eta, &xtol, &w[0], &iflag);
      if (iflag < 0) {
	strcpy (task, "LBFGS: routine stops with unexpected error.");
	return -1;
      } 
      if (iflag == 0) {
	strcpy (task, "LBFGS: converge");
	return 0; // terminate
      } 
      return 1; // evaluate next f and g
    }
      break;
      
    case 1: {
      long iprint = -1;
      double factr = 1.0; // may be enough
      double pgtol = 0;
      if (iflag == 0) strcpy (task, "START"); // init
      iflag = 1;

      while (true) {
	lbfgsb (&n, &m, x, l, u, nbd, f, g, &factr, &pgtol, &w[0], &iw[0], 
		task, &iprint, csave, lsave, isave, dsave, 60, 60);
	if (strncmp (task, "FG", 2) == 0) return 1;         // evaluate next f and g
	else if (strncmp (task, "NEW_X", 4) == 0) continue; // next iteration
	else if (strncmp (task, "CONV",  4) == 0) return 0; // terminate
	else return -1; // error
      }

      return 1;
    }
      break;

    default:
      strcpy (task, "LBFGS::run() call LBFGS::init() first");
      return -1;
    }

    return -1;
  }
};
